from .database import curvedb
from .plot import curvedrawer

class curve(curvedb, curvedrawer):
    def __init__(self, db_path='', db_name='curvedb', figsize=(15, 12)):
        curvedb.__init__(self, db_path=db_path, db_name=db_name)
        curvedrawer.__init__(self, figsize=figsize)

    def show(self):
        curvedb.show(self)
        curvedrawer.show(self)
    
    def log(self, dict_value):
        self.add_value(dict_value)
    
    def draw(self, attr_x, attr_y, label=None):
        self.drawcurves(self.get_column(attr_x), self.get_column(attr_y), label)

    def daemon(self, start=False, interval=1):
        if start:
            assert interval >= 1
            self.daemon_counter = 0
            self.daemon_interval = interval
        else:
            if self.daemon_counter == self.daemon_interval:
                self.daemon_counter = 1
                return True
            else:
                self.daemon_counter += 1
        return False
        
class losscurve(curve):
    def __init__(self, db_path, db_name, load_iter=False, figsize=(15, 12)):
        curve.__init__(self, db_path=db_path, db_name=db_name, figsize=figsize)
        self.iteration = 0
        self.stepsize = 1
        self.metadata = curvedb(db_path, 'metadata')
        self.states = curvedb(db_path, 'states')
        if load_iter:
            if self.states.check_index(0):
                if 'cur_iter' in self.states.get_value(0):
                    self.iteration = self.states.get_value(0)['cur_iter']
        self.internal_iter = False

    def set_iter(self, iteration):
        self.iteration = iteration

    def set_step(self, stepsize):
        self.stepsize = stepsize

    def set_internal(self, state=True):
        self.internal_iter = state
        if self.internal_iter:
            self.set_xlabel('iteration')

    def step(self, no_step=False):
        self.metadata.add_value({'iteration': self.iteration})
        if not no_step:
            self.iteration += self.stepsize
            if self.states.check_index(0):
                states = self.states.get_value(0)
                states['cur_iter'] = self.iteration
                self.states.modify_value(0, states)
            else:
                self.states.add_value({'cur_iter': self.iteration})
    
    def log(self, dict_value, no_step=False):
        curve.log(self, dict_value)
        self.step(no_step=no_step)

    def reset(self, force_clean=False):
        curve.reset(self, force_clean=force_clean)
        self.metadata.reset(force_clean=force_clean)
        self.states.reset(force_clean=force_clean)
        self.iteration = 0

    def show(self):
        curvedb.show(self)
        self.metadata.show()
        self.states.show()
        curvedrawer.show(self)
        
    def draw_internal(self, attr=None, label=None):
        self.drawcurves(self.metadata.get_column('iteration'), self.get_column(attr), label)

    def draw_external(self, attr_x=None, attr_y=None, label=None):
        curve.draw(self, attr_x, attr_y, label)
        
    def draw(self, *args, **kwargs):
        if not self.internal_iter:
            self.draw_external(*args, **kwargs)
        else:
            self.draw_internal(*args, **kwargs)

    def synchronize(self):
        curve.save(self)
        self.metadata.save()
        self.states.save()

if __name__ == '__main__':
    import time
    import ipdb
    import numpy as np
    from tqdm import tqdm
    curve_handler = losscurve(db_path='debug', db_name='data', figsize=(20, 12))
    curve_handler.reset(True)
    curve_handler.set_internal()
    curve_handler.set_ylabel('loss', False)
    curve_handler.set_ylabel('acc', True)
    curve_handler.daemon(True, 10)
    for iteration in tqdm(range(501)):
        curve_handler.clean()
        curve_handler.log({'loss_A': 0.11*np.exp(-iteration/500), 'loss_B': 0.11*np.exp(-iteration/1000)}, no_step=iteration % 10 == 0)
        if iteration % 10 == 0:
            curve_handler.log({'acc_A': 1 / (1 + np.exp(-iteration/500)), 'acc_B': 1 / (1 + np.exp(-iteration/1000))})
        
        to_draw = curve_handler.daemon()
        if to_draw:
            curve_handler.draw('loss_A', 'model A')
            curve_handler.draw('loss_B', 'model B')
            curve_handler.twin()
            curve_handler.clean()
            curve_handler.draw('acc_A', 'model A')
            curve_handler.draw('acc_B', 'model B')
            curve_handler.twin()
            curve_handler.reset_choice()
            curve_handler.legend(inside=False)
            curve_handler.synchronize()
            curve_handler.render('test.png')
    


import pandas as pd
import os
import numpy as np
import lmdb
import pickle
import base64
def obj2bytes(obj):
    return base64.b64encode(pickle.dumps(obj))
def str2obj(str):
    return pickle.loads(base64.b64decode(str))

class curvedb(object):
    '''
    database class for storing curve data.
    '''
    def __init__(self, db_path=None, db_name='curvedb'):
        self.db_name = db_name
        self.db_path = db_path
        assert self.db_path is not None
        assert len(self.db_path) > 0
        self.initialize()
        
    def initialize(self):
        try:
            self.load()
        except:
            self.db = pd.DataFrame()
            self.save()
        
    def reset(self, force_clean=False):
        self.db = pd.DataFrame()
        if force_clean:
            self.save()

    def get_db(self):
        return self.db

    @property
    def keys(self):
        return list(self.db.keys())
    
    @property
    def cursor(self):
        return len(self.db)

    def get_column(self, key):
        return self.db[key].to_numpy()

    def add_key(self, key):
        if isinstance(key, list):
            for each_key in key:
                self.db[each_key] = np.nan
        else:
            self.db[key] = np.nan

    def delete_key(self, key):
        if isinstance(key, list):
            self.db.drop(columns=key, inplace=True)
        else:
            self.db.drop(columns=[key], inplace=True)
            
    def search_value(self, key, target_value):
        return self.db[key] == target_value

    def check_index(self, index):
        return index < len(self.db)

    def get_value(self, index):
        return self.db.loc[index]

    def add_value(self, dict_value):
        assert isinstance(dict_value, dict), 'The input of add_value should be a dict variable!'

        for key in dict_value:
            if key not in self.keys:
                self.add_key(key)

        for key in self.keys:
            if key not in dict_value:
                dict_value[key] = np.nan

        self.db = self.db.append([dict_value], ignore_index=True)

    def modify_value(self, index, dict_value):
        self.db.loc[index] = dict_value

    def del_value(self, index):
        if isinstance(key, list):
            self.db.drop(index=index)
        else:
            self.db.drop(index=[index])
    
    def show(self):
        print(self.db)

    def save(self, path=''):
        if len(path) > 0:
            self.db_path = path
        env = lmdb.open(self.db_path, map_size=int(1e12))
        txn = env.begin(write=True)
        txn.put(key=self.db_name.encode(), value=obj2bytes(self.db))
        txn.commit()
        env.close()
        return

    def load(self):
        env = lmdb.open(self.db_path, map_size=int(1e12))
        txn = env.begin(write=False)
        self.db = str2obj(txn.get(self.db_name.encode()).decode())
        env.close()
        return


from matplotlib import pyplot as plt
import matplotlib
matplotlib.use('Agg')
import numpy as np
import itertools
import os
import random
import warnings
warnings.filterwarnings('ignore')

class curvedrawer(object):
    def __init__(self, figsize=(15, 12)):
        self.colorbank = ['red', 'blue', 'green', 'purple', 'brown', 'gray', 'orange', 'olive', 'cyan', 'pink']
        self.linestyle = ['solid', 'dotted', 'dashed', 'dashdot']
        self.pointname = ['point', 'triangle_down', 'triangle_up', 'octagon', 'square', 'pentagon', 'star', 'hexagon1', 'hexagon2', 'cross', 'diamond', 'circle', 'plus']
        self.pointstyle = ['.', 'v', '^', '8', 's', 'p', '*', 'h', 'H', 'x', 'D', 'o', '+']
        assert len(self.pointname) == len(self.pointstyle)
        self.pointdict = dict()
        for point_idx, pointname in enumerate(self.pointname):
            self.pointdict[pointname] = self.pointstyle[point_idx]
        self.colorbank_num = len(self.colorbank)
        self.linestyle_num = len(self.linestyle)
        self.pointname_num = len(self.pointname)
        self.choice_num = self.colorbank_num * self.linestyle_num * self.pointname_num
        self.choices = list(itertools.product(self.pointname, self.linestyle, self.colorbank))
        self.choice_cursor = 0
        self.panel = plt.figure(figsize=figsize)
        self.twin_state = False
        self.ori_figure = self.panel.add_subplot(111)
        self.figsize = figsize
        self.twin_figure = None
        self.x_label = None
        self.y_label = None
        self.y_twin_label = None

    def shuffle(self):
        random.shuffle(self.choices)
    
    def reset_choice(self):
        self.choice_cursor = 0

    def set_xlabel(self, x_name=None):
        if x_name is not None:
            self.ori_figure.set_xlabel(x_name)
            self.x_label = x_name

    def set_ylabel(self, y_name=None, twin_state=False):
        if y_name is not None:
            if twin_state:
                if self.twin_figure is None:
                    self.twin_figure = self.ori_figure.twinx()
                self.twin_figure.set_ylabel(y_name)
                self.y_twin_label = y_name
            else:
                self.ori_figure.set_ylabel(y_name)
                self.y_label = y_name

    def drawcurve(self, x, y, color='blue', linestyle='solid', pointstyle='point', y_label=None): 
        x_to_draw = np.array(x)
        y_to_draw = np.array(y)
        assert x_to_draw.ndim == 1, 'x should be an 1D array!'
        assert y_to_draw.ndim == 1, 'y should be an 1D array!'
        assert color in self.colorbank, 'color is not in internal color bank!'
        assert linestyle in self.linestyle, 'linestyle is not in internal linestyle bank!'
        assert pointstyle in self.pointname, 'pointstyle is not in internal pointstyle bank!'
        assert len(y_to_draw) == len(x_to_draw), 'the length of x is not equal to the length of y!'
        valid = np.logical_and(np.logical_not(np.isnan(x_to_draw)), np.logical_not(np.isnan(y_to_draw)))
        x_to_draw = x_to_draw[valid]
        y_to_draw = y_to_draw[valid]
        
        if self.twin_state:
            self.twin_figure.plot(x_to_draw, y_to_draw, color=color, linestyle=linestyle, marker=self.pointdict[pointstyle], label=y_label)
        else:
            self.ori_figure.plot(x_to_draw, y_to_draw, color=color, linestyle=linestyle, marker=self.pointdict[pointstyle], label=y_label)
        return

    def drawcurves(self, x, y, y_labels=None):
        x_to_draw = np.array(x)
        y_to_draw = np.array(y)
        
        if y_to_draw.ndim == 1:
            y_to_draw = np.expand_dims(y_to_draw, axis=1)
        assert x_to_draw.ndim == 1, 'x should be an 1D array!'
        if isinstance(y_labels, str):
            y_labels = [y_labels]
        if y_labels is not None:
            assert len(y_labels) == y_to_draw.shape[1], 'the length of labels should be equal to the number of curves!'
        assert y_to_draw.shape[1] <= self.choice_num, 'anycurve now only supports {:d} types of curve while the number of input curves is {:d}.'.format(self.choice_num, y_to_draw.shape[1])
        y_cursor = 0
        for pointstyle, linestyle, color in self.choices[self.choice_cursor:]:
            y_label_used = None if y_labels is None else y_labels[y_cursor]
            self.drawcurve(x_to_draw, y_to_draw[:, y_cursor], color=color, linestyle=linestyle, pointstyle=pointstyle, y_label=y_label_used)
            self.choice_cursor += 1
            y_cursor += 1
            if y_cursor == y_to_draw.shape[1]:
                break

    def clean(self):
        if self.twin_state:
            self.twin_figure.cla()
        else:
            self.ori_figure.cla() 
        self.set_xlabel(self.x_label)
        self.set_ylabel(self.y_label, False)
        if self.y_twin_label is not None:
            self.set_ylabel(self.y_twin_label, True)
            
    def legend(self, inside=True, left_offset=None, right_offset=None):
        if inside:
            if left_offset is None:
                left_offset = (0, 1)
            if right_offset is None:
                right_offset = (1, 1)

            self.ori_figure.legend(bbox_to_anchor=left_offset, loc=2)
            if self.twin_figure is not None:
                self.twin_figure.legend(bbox_to_anchor=right_offset, loc=1)
        else:
            if left_offset is None:
                left_offset = (-0.12, 1.05)
            if right_offset is None:
                right_offset = (1.12, 1.05)

            self.ori_figure.legend(bbox_to_anchor=left_offset, loc=2)
            if self.twin_figure is not None:
                self.twin_figure.legend(bbox_to_anchor=right_offset, loc=1)

    def show(self):
        plt.show()
    
    def twin(self):
        if not self.twin_state:
            if self.twin_figure is None:
                self.twin_figure = self.ori_figure.twinx()
            self.twin_state = True
        else:
            self.twin_state = False

    def render(self, filename):
        dirname = os.path.dirname(os.path.abspath(filename))
        os.makedirs(dirname, exist_ok=True)
        self.ori_figure.figure.savefig(filename)

if __name__ == '__main__':
    a = curvedrawer((20, 12))
    # a.drawcurve([1, 2], [3, 4], 'red', 'solid', 'point', 'x_axis', 'y_axis', 'good')

    a.drawcurves([1, 2, 3], [[3,5,7], [4,6,8], [5,7,9]], ['good', 'bad', 'ugly'])
    a.twin()
    a.drawcurves([1, 2, 3], [[3,1,-1], [2,0,-2], [1,-1,-3]], ['good', 'bad', 'ugly'])
    a.twin()
    a.drawcurves([1, 2, 3], [[4,6,8], [5,7,9], [6,8,10]], ['good', 'bad', 'ugly'])
    a.twin()
    a.drawcurves([1, 2, 3], [[2,0,-2], [1,-1,-3], [0,-2,-4]], ['good', 'bad', 'ugly'])

    # a.drawcurves([1, 2, 3], [[3,5], [4,6], [5,7]], 'x_axis', 'y_axis', ['good', 'bad'])
    # a.twin()
    # a.drawcurves([1, 2, 3], [[3,1], [2,0], [1,-1]], 'x_axis', 'another_y_axis', ['good_twin', 'bad_twin'])
    # a.twin()
    # a.drawcurves([1, 2, 3], [[4,6], [5,7], [6,8]], 'x_axis', 'y_axis', ['good2', 'bad2'])
    # a.twin()
    # a.drawcurves([1, 2, 3], [[2,0], [1,-1], [0,-2]], 'x_axis', 'another_y_axis', ['good2_twin', 'bad2_twin'])
    # a.twin()

    a.legend(inside=True)
    a.show()
    a.save('test.png')


from setuptools import setup, find_packages
setup(
    name = "anycurve",
    version = "0.3.0",
    keywords = ["anycurve", "curve", "plot", "hwfan"],
    description = "A data visualization toolbox",
    long_description = "A python toolbox for the visualization of data series",
    license = "Apache",

    url = "http://hwfan.io",
    author = "hwfan",
    author_email = "hwnorm@outlook.com",

    packages = find_packages(),
    include_package_data = True,
    platforms = "any",
    install_requires = ['pandas', 'numpy', 'lmdb', 'matplotlib'],
    scripts = [],

)
